from urllib.parse import urlencode


class QueryStringTransformer:
    """
    A utility class to manipulate query strings.

    Args:
        args: An object representing the query parameters, typically an
              ImmutableMultiDict (Django) or QueryDict (Flask) which can be
              accessed with request.GET (Django) or request.args (Flask).
    """

    def __init__(self, args=None) -> None:
        if isinstance(args, list):
            self.args = args
        elif args is not None:
            try:
                args_lists = args.lists()
            except AttributeError:
                raise AttributeError(
                    "args must be an ImmutableMultiDict (Django), a QueryDict (Flask) object or an iterable of (key, [values]) tuples"
                )
            self.args = list(args_lists)
        else:
            self.args = []

    def parameter_exists(self, parameter) -> bool:
        """
        Check if a parameter exists in the query parameters.
        """

        for key, _ in self.args:
            if key == parameter:
                return True
        return False

    def parameter_values(self, parameter: str) -> list:
        """
        Get the values associated with a parameter in the query parameters.
        Raises a KeyError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                return values
        raise KeyError(f"Parameter '{parameter}' does not exist")

    def add_parameter(
        self, parameter: str, values: str | int | list | None = None
    ) -> "QueryStringTransformer":
        """
        Add a new parameter to the query parameters.
        Raises a ValueError if the parameter already exists.
        """

        for key, vals in self.args:
            if key == parameter:
                raise ValueError(f"Parameter '{parameter}' already exists")
        if not isinstance(values, list):
            values = [str(values)] if values is not None else []
        else:
            values = [str(v) for v in values]
        self.args.append((parameter, values))
        return self

    def update_parameter(
        self, parameter: str, values: str | int | list | None = None
    ) -> "QueryStringTransformer":
        """
        Update an existing parameter in the query parameters.
        If the parameter does not exist, it will be added.
        """

        if self.parameter_exists(parameter):
            self.remove_parameter(parameter)
        self.add_parameter(parameter, values)
        return self

    def remove_parameter(self, parameter: str) -> "QueryStringTransformer":
        """
        Remove a parameter from the query parameters.
        Raises a KeyError if the parameter does not exist.
        """

        for index, (key, vals) in enumerate(self.args):
            if key == parameter:
                del self.args[index]
                return self
        raise KeyError(f"Parameter '{parameter}' does not exist")

    def is_value_in_parameter(self, parameter: str, value: str | int) -> bool:
        """
        Check if a specific value exists within a parameter's values.
        Raises a KeyError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                return str(value) in values
        raise KeyError(f"Parameter '{parameter}' does not exist")

    def add_parameter_value(
        self, parameter: str, value: str | int
    ) -> "QueryStringTransformer":
        """
        Add a specific value to a parameter's values.
        Raises a KeyError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                if str(value) not in values:
                    values.append(str(value))
                return self
        raise KeyError(f"Parameter '{parameter}' does not exist")

    def toggle_parameter_value(
        self, parameter: str, value: str | int
    ) -> "QueryStringTransformer":
        """
        Toggle a value within a parameter's values.
        If the value exists, it will be removed; if it does not exist, it will be added.
        Raises a KeyError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                str_value = str(value)
                if str_value in values:
                    values.remove(str_value)
                else:
                    if str_value not in values:
                        values.append(str_value)
                return self
        raise KeyError(f"Parameter '{parameter}' does not exist")

    def remove_parameter_value(
        self, parameter: str, value: str | int
    ) -> "QueryStringTransformer":
        """
        Remove a specific value from a parameter's values.
        Raises a KeyError if the parameter does not exist or if the value is not present.
        """

        for key, values in self.args:
            if key == parameter:
                if str(value) in values:
                    values.remove(str(value))
                    return self
                raise KeyError(
                    f"Value '{value}' does not exist for parameter '{parameter}'"
                )
        raise KeyError(f"Parameter '{parameter}' does not exist")

    def get_query_string(self) -> str:
        """
        Get the full query string.
        Returns an empty string if there are no query parameters.
        """

        query = urlencode(self.args, doseq=True)
        if not query:
            return ""
        return f"?{query}"



import unittest

from tna_utilities.number import numberish


class TestNumber(unittest.TestCase):

    def test_happy(self):
        self.assertEqual(numberish(0), "None")
        self.assertEqual(numberish(1), "1")
        self.assertEqual(numberish(99), "99")
        self.assertEqual(numberish(100), "100")
        self.assertEqual(numberish(999), "999")
        self.assertEqual(numberish(1000), "1 thousand")
        self.assertEqual(numberish(1001), "About 1 thousand")
        self.assertEqual(numberish(1337), "About 1.3 thousand")
        self.assertEqual(numberish(9999), "About 10 thousand")
        self.assertEqual(numberish(10000), "10 thousand")
        self.assertEqual(numberish(10001), "About 10 thousand")
        self.assertEqual(numberish(11000), "11 thousand")
        self.assertEqual(numberish(11001), "About 11 thousand")
        self.assertEqual(numberish(11499), "About 11 thousand")
        self.assertEqual(numberish(11500), "About 12 thousand")
        self.assertEqual(numberish(11999), "About 12 thousand")
        self.assertEqual(numberish(100000), "100 thousand")
        self.assertEqual(numberish(100001), "About 100 thousand")
        self.assertEqual(numberish(1000000), "1 million")
        self.assertEqual(numberish(1000001), "About 1 million")
        self.assertEqual(numberish(1234567), "About 1.2 million")
        self.assertEqual(numberish(12345678), "About 12 million")
        self.assertEqual(numberish(123456789), "About 120 million")
        self.assertEqual(numberish(1234567890), "About 1.2 billion")
        self.assertEqual(numberish(1337.1337), "About 1.3 thousand")

    def test_happy_simplified(self):
        self.assertEqual(numberish(0, simple_units=True), "None")
        self.assertEqual(numberish(1, simple_units=True), "1")
        self.assertEqual(numberish(99, simple_units=True), "99")
        self.assertEqual(numberish(100, simple_units=True), "100")
        self.assertEqual(numberish(999, simple_units=True), "999")
        self.assertEqual(numberish(1000, simple_units=True), "1k")
        self.assertEqual(numberish(1001, simple_units=True), "About 1k")
        self.assertEqual(numberish(1337, simple_units=True), "About 1.3k")
        self.assertEqual(numberish(9999, simple_units=True), "About 10k")
        self.assertEqual(numberish(10000, simple_units=True), "10k")
        self.assertEqual(numberish(10001, simple_units=True), "About 10k")
        self.assertEqual(numberish(11000, simple_units=True), "11k")
        self.assertEqual(numberish(11001, simple_units=True), "About 11k")
        self.assertEqual(numberish(11499, simple_units=True), "About 11k")
        self.assertEqual(numberish(11500, simple_units=True), "About 12k")
        self.assertEqual(numberish(11999, simple_units=True), "About 12k")
        self.assertEqual(numberish(100000, simple_units=True), "100k")
        self.assertEqual(numberish(100001, simple_units=True), "About 100k")
        self.assertEqual(numberish(1000000, simple_units=True), "1m")
        self.assertEqual(numberish(1000001, simple_units=True), "About 1m")
        self.assertEqual(numberish(1234567, simple_units=True), "About 1.2m")
        self.assertEqual(numberish(12345678, simple_units=True), "About 12m")
        self.assertEqual(numberish(123456789, simple_units=True), "About 120m")
        self.assertEqual(numberish(1234567890, simple_units=True), "About 1.2b")
        self.assertEqual(numberish(1337.1337, simple_units=True), "About 1.3k")

    def test_happy_prefix_string(self):
        self.assertEqual(
            numberish(1499999, simple_units=True, prefix_text="Approx "), "Approx 1.5m"
        )
        self.assertEqual(
            numberish(1499999, simple_units=True, prefix_text="~"), "~1.5m"
        )

    def test_happy_prefix_tuple(self):
        self.assertEqual(
            numberish(1000000, simple_units=True, prefix_text=("Almost ", "Over ")),
            "1m",
        )
        self.assertEqual(
            numberish(1499999, simple_units=True, prefix_text=("Almost ", "Over ")),
            "Almost 1.5m",
        )
        self.assertEqual(
            numberish(1500000, simple_units=True, prefix_text=("Almost ", "Over ")),
            "1.5m",
        )
        self.assertEqual(
            numberish(1500001, simple_units=True, prefix_text=("Almost ", "Over ")),
            "Over 1.5m",
        )

    def test_unhappy(self):
        with self.assertRaises(ValueError):
            numberish("one")
        with self.assertRaises(ValueError):
            numberish(None)
        with self.assertRaises(ValueError):
            numberish({})
        with self.assertRaises(ValueError):
            numberish([])



import math
from typing import Union


def numberish(
    value: Union[float, int],
    simple_units: bool = False,
    prefix_text: Union[str, tuple[str, str]] = "About ",
) -> str:
    """
    Convert a number into a human-readable string with appropriate units.
    """

    if not isinstance(value, (int, float)):
        raise ValueError("Value must be an integer or float")
    if value == 0:
        return "None"
    units = [
        ("billion", "b", 1_000_000_000),
        ("million", "m", 1_000_000),
        ("thousand", "k", 1_000),
    ]
    for unit_full, unit_simple, threshold in units:
        if value >= threshold:
            unit = unit_simple if simple_units else " " + unit_full
            base_value = value / threshold
            if base_value.is_integer():
                return f"{int(base_value)}{unit}"
            base_value_rounded = round(
                base_value,
                -int(math.floor(math.log10(abs(base_value)))) + 1,
            )
            if base_value_rounded == base_value:
                prefix_text = ""
            elif isinstance(prefix_text, tuple):
                if len(prefix_text) != 2:
                    raise ValueError("prefix_text tuple must have exactly two elements")
                if not all(isinstance(pt, str) for pt in prefix_text):
                    raise ValueError(
                        "Both elements of the prefix_text tuple must be strings"
                    )
                approximation_is_over = base_value_rounded * threshold - value >= 0
                prefix_text = (
                    prefix_text[0] if approximation_is_over else prefix_text[1]
                )
            return f"{prefix_text}{base_value_rounded:g}{unit}"
    return str(int(value))
